# -*- coding: utf-8 -*-


"""Top-level package for CID (Content IDentifier)."""

__author__ = """Dhruv Baldawa"""
__email__ = "dhruv@dhruvb.com"
__version__ = "0.4.2"

from .cid import CIDv0, CIDv1, make_cid, is_cid, from_string, from_bytes  # noqa: F401



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "base58",
    "py-multibase",
    "py-multicodec",
    "morphys",
    "planetmint-multiformat-multihash",
]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
    "pytest-cov",
    "hypothesis",
]

setup(
    name="planetmint-py-cid",
    version="0.4.2",
    description="Self-describing content-addressed identifiers for distributed systems",
    long_description=readme + "\n\n" + history,
    author="Dhruv Baldawa",
    author_email="mail@planetmint.io",
    url="https://github.com/planetmint/py-cid",
    packages=find_packages(include=["cid"]),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords="cid",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite="tests",
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



# -*- coding: utf-8 -*-
import base58

import multibase

import multihash as mh

from morphys import ensure_bytes, ensure_unicode

import multicodec


class BaseCID(object):
    __hash__ = object.__hash__

    def __init__(self, version, codec, multihash):
        """
        Creates a new CID object. This class should not be used directly, use :py:class:`cid.cid.CIDv0` or
        :py:class:`cid.cid.CIDv1` instead.


        :param int version: CID version (0 or 1)
        :param str codec: codec to be used for encoding the hash
        :param str multihash: the multihash
        """
        self._version = version
        self._codec = codec
        self._multihash = ensure_bytes(multihash)

    @property
    def version(self):
        """CID version"""
        return self._version

    @property
    def codec(self):
        """CID codec"""
        return self._codec

    @property
    def multihash(self):
        """CID multihash"""
        return self._multihash

    @property
    def buffer(self):
        raise NotImplementedError

    def encode(self, *args, **kwargs):
        raise NotImplementedError

    def __repr__(self):
        def truncate(s, length):
            return s[:length] + b".." if len(s) > length else s

        truncate_length = 20
        return "{class_}(version={version}, codec={codec}, multihash={multihash})".format(
            class_=self.__class__.__name__,
            version=self._version,
            codec=self._codec,
            multihash=truncate(self._multihash, truncate_length),
        )

    def __str__(self):
        return ensure_unicode(self.encode())

    def __eq__(self, other):
        return (self.version == other.version) and (self.codec == other.codec) and (self.multihash == other.multihash)


class CIDv0(BaseCID):
    """CID version 0 object"""

    CODEC = "dag-pb"

    def __init__(self, multihash):
        """
        :param bytes multihash: multihash for the CID
        """
        super(CIDv0, self).__init__(0, self.CODEC, multihash)

    @property
    def buffer(self):
        """
        The raw representation that will be encoded.

        :return: the multihash
        :rtype: bytes
        """
        return self.multihash

    def encode(self):
        """
        base58-encoded buffer

        :return: encoded representation or CID
        :rtype: bytes
        """
        return ensure_bytes(base58.b58encode(self.buffer))

    def to_v1(self):
        """
        Get an equivalent :py:class:`cid.CIDv1` object.

        :return: :py:class:`cid.CIDv1` object
        :rtype: :py:class:`cid.CIDv1`
        """
        return CIDv1(self.CODEC, self.multihash)


class CIDv1(BaseCID):
    """CID version 1 object"""

    def __init__(self, codec, multihash):
        super(CIDv1, self).__init__(1, codec, multihash)

    @property
    def buffer(self):
        """
        The raw representation of the CID

        :return: raw representation of the CID
        :rtype: bytes
        """
        return b"".join([bytes([self.version]), multicodec.add_prefix(self.codec, self.multihash)])

    def encode(self, encoding="base58btc"):
        """
        Encoded version of the raw representation

        :param str encoding: the encoding to use to encode the raw representation, should be supported by
            ``py-multibase``
        :return: encoded raw representation with the given encoding
        :rtype: bytes
        """
        return multibase.encode(encoding, self.buffer)

    def to_v0(self):
        """
        Get an equivalent :py:class:`cid.CIDv0` object.

        :return: :py:class:`cid.CIDv0` object
        :rtype: :py:class:`cid.CIDv0`
        :raise ValueError: if the codec is not 'dag-pb'
        """
        if self.codec != CIDv0.CODEC:
            raise ValueError("CIDv1 can only be converted for codec {}".format(CIDv0.CODEC))

        return CIDv0(self.multihash)


def make_cid(*args):
    """
    Creates a :py:class:`cid.CIDv0` or :py:class:`cid.CIDv1` object based on the given parameters

    The function supports the following signatures:

            make_cid(<base58 encoded multihash CID>) -> CIDv0

            make_cid(<multihash CID>) -> CIDv0

            make_cid(<multibase encoded multihash CID>) -> CIDv1

            make_cid(<version>, <codec>, <multihash>) -> CIDv1

    :param args:
        - base58-encoded multihash (str or bytes)
        - multihash (str or bytes)
        - multibase-encoded multihash (str or bytes)
        - version:int, codec(str), multihash(str or bytes)
    :returns: the respective CID object
    :rtype: :py:class:`cid.CIDv0` or :py:class:`cid.CIDv1`
    :raises ValueError: if the number of arguments is not 1 or 3
    :raises ValueError: if the only argument passed is not a ``str`` or a ``byte``
    :raises ValueError: if the string provided is not a valid base58 encoded hash
    :raises ValueError: if 3 arguments are passed and version is not 0 or 1
    :raises ValueError: if 3 arguments are passed and the ``codec`` is not supported by ``multicodec``
    :raises ValueError: if 3 arguments are passed and the ``multihash`` is not ``str`` or ``byte``
    :raises ValueError: if 3 arguments are passed with version 0 and codec is not *dag-pb*
    """
    if len(args) == 1:
        data = args[0]
        if isinstance(data, str):
            return from_string(data)
        elif isinstance(data, bytes):
            return from_bytes(data)
        else:
            raise ValueError("invalid argument passed, expected: str or byte, found: {}".format(type(data)))

    elif len(args) == 3:
        version, codec, multihash = args
        if version not in (0, 1):
            raise ValueError("version should be 0 or 1, {} was provided".format(version))
        if not multicodec.is_codec(codec):
            raise ValueError("invalid codec {} provided, please check".format(codec))
        if not (isinstance(multihash, str) or isinstance(multihash, bytes)):
            raise ValueError("invalid type for multihash provided, should be str or bytes")

        if version == 0:
            if codec != CIDv0.CODEC:
                raise ValueError("codec for version 0 can only be {}, found: {}".format(CIDv0.CODEC, codec))
            return CIDv0(multihash)
        else:
            return CIDv1(codec, multihash)
    else:
        raise ValueError("invalid number of arguments, expected 1 or 3")


def is_cid(cidstr):
    """
    Checks if a given input string is valid encoded CID or not.
    It takes same input as `cid.make_cid` method with a single argument


    :param cidstr: input string which can be a

        - base58-encoded multihash
        - multihash
        - multibase-encoded multihash
    :type cidstr: str or bytes
    :return: if the value is a valid CID or not
    :rtype: bool
    """
    try:
        return bool(make_cid(cidstr))
    except ValueError:
        return False


def from_string(cidstr):
    """
    Creates a CID object from a encoded form

    :param str cidstr: can be

        - base58-encoded multihash
        - multihash
        - multibase-encoded multihash
    :return: a CID object
    :rtype: :py:class:`cid.CIDv0` or :py:class:`cid.CIDv1`
    """
    cidbytes = ensure_bytes(cidstr, "utf-8")
    return from_bytes(cidbytes)


def from_bytes(cidbytes):
    """
    Creates a CID object from a encoded form

    :param bytes cidbytes: can be

        - base58-encoded multihash
        - multihash
        - multibase-encoded multihash
    :return: a CID object
    :rtype: :py:class:`cid.CIDv0` or :py:class:`cid.CIDv1`
    :raises: `ValueError` if the base58-encoded string is not a valid string
    :raises: `ValueError` if the length of the argument is zero
    :raises: `ValueError` if the length of decoded CID is invalid
    """
    if len(cidbytes) < 2:
        raise ValueError("argument length can not be zero")

    # first byte for identity multibase and CIDv0 is 0x00
    # putting in assumption that multibase for CIDv0 can not be identity
    # refer: https://github.com/ipld/cid/issues/13#issuecomment-326490275
    if cidbytes[0] != 0 and multibase.is_encoded(cidbytes):
        # if the bytestream is multibase encoded
        cid = multibase.decode(cidbytes)

        if len(cid) < 2:
            raise ValueError("cid length is invalid")

        data = cid[1:]
        version = int(cid[0])
        codec = multicodec.get_codec(data)
        multihash = multicodec.remove_prefix(data)
    elif cidbytes[0] in (0, 1):
        # if the bytestream is a CID
        version = cidbytes[0]
        data = cidbytes[1:]
        codec = multicodec.get_codec(data)
        multihash = multicodec.remove_prefix(data)
    else:
        # otherwise its just base58-encoded multihash
        try:
            version = 0
            codec = CIDv0.CODEC
            multihash = base58.b58decode(cidbytes)
        except ValueError:
            raise ValueError("multihash is not a valid base58 encoded multihash")

    try:
        mh.decode(multihash)
    except ValueError:
        raise

    return make_cid(version, codec, multihash)



import hashlib
import multihash
import pytest
import base58
import string
from morphys import ensure_unicode

import multibase
import multicodec
from cid import CIDv0, CIDv1, make_cid, is_cid, from_string
from multibase.multibase import ENCODINGS
from hypothesis import given, strategies as st


ALLOWED_ENCODINGS = [encoding for encoding in ENCODINGS if encoding.code != b"\x00"]


@pytest.fixture(scope="session")
def test_hash():
    data = hashlib.sha256(b"hello world").hexdigest()
    return multihash.encode(bytes.fromhex(data), "sha2-256")


class CIDv0TestCase(object):
    @pytest.fixture()
    def cid(self, test_hash):
        return CIDv0(test_hash)

    def test_init(self, cid, test_hash):
        """#__init__: initializes all the attributes correctly"""
        assert cid.version == 0
        assert cid.codec == CIDv0.CODEC
        assert cid.multihash == test_hash

    def test_buffer(self, cid, test_hash):
        """.buffer: is computed properly"""
        assert cid.buffer == test_hash

    def test_encode(self, cid):
        """#encode: base58 encodes the buffer by default"""
        assert cid.encode() == b"QmaozNR7DZHQK1ZcU9p7QdrshMvXqWK6gpu5rmrkPdT3L4"

    def test_str(self, cid):
        assert str(cid) == ensure_unicode(cid.encode())


class CIDv1TestCase(object):
    TEST_CODEC = "dag-pb"

    @pytest.fixture()
    def cid(self, test_hash):
        return CIDv1(self.TEST_CODEC, test_hash)

    def test_init(self, cid, test_hash):
        """#__init__: attributes are set properly"""
        assert cid.version == 1
        assert cid.codec == self.TEST_CODEC
        assert cid.multihash == test_hash

    def test_buffer(self, cid):
        """.buffer: buffer is computed properly"""
        buffer = cid.buffer
        assert buffer[0] == 1
        assert buffer[1:] == multicodec.add_prefix(self.TEST_CODEC, cid.multihash)

    def test_encode_default(self, cid):
        """#encode defaults to base58btc encoding"""
        assert cid.encode() == b"zdj7WhuEjrB52m1BisYCtmjH1hSKa7yZ3jEZ9JcXaFRD51wVz"

    @pytest.mark.parametrize("codec", ENCODINGS)
    def test_encode_encoding(self, cid, codec):
        """#encode uses the encoding provided for encoding"""
        assert cid.encode(codec.encoding) == multibase.encode(codec.encoding, cid.buffer)

    def test_str(self, cid):
        assert str(cid) == ensure_unicode(cid.encode())


class CIDTestCase(object):
    def test_cidv0_eq_cidv0(self, test_hash):
        """check for equality for CIDv0 for same hash"""
        assert CIDv0(test_hash) == make_cid(CIDv0(test_hash).encode())

    def test_cidv0_neq(self):
        """check for inequality for CIDv0 for different hashes"""
        assert CIDv0(b"QmdfTbBqBPQ7VNxZEYEj14VmRuZBkqFbiwReogJgS1zR1n") != CIDv0(
            b"QmdfTbBqBPQ7VNxZEYEj14VmRuZBkqFbiwReogJgS1zR1o"
        )

    def test_cidv0_eq_cidv1(self, test_hash):
        """check for equality between converted v0 to v1"""
        assert CIDv0(test_hash).to_v1() == CIDv1(CIDv0.CODEC, test_hash)

    def test_cidv1_eq_cidv0(self, test_hash):
        """check for equality between converted v1 to v0"""
        assert CIDv1(CIDv0.CODEC, test_hash).to_v0() == CIDv0(test_hash)

    def test_cidv1_to_cidv0_no_dag_pb(self):
        """converting non dag-pb CIDv1 should raise an exception"""
        with pytest.raises(ValueError) as excinfo:
            CIDv1("base2", test_hash).to_v0()
        assert "can only be converted for codec" in str(excinfo.value)

    @pytest.mark.parametrize(
        "test_cidv0",
        ("QmbWqxBEKC3P8tqsKc98xmWNzrzDtRLMiMPL8wBuTGsMnR", "QmUNLLsPACCz1vLxQVkXqqLX5R1X345qqfHbsf67hvA3Nn"),
    )
    def test_is_cidv0_valid(self, test_cidv0):
        assert is_cid(test_cidv0)
        assert is_cid(make_cid(test_cidv0).encode())

    @given(hash=st.text(string.ascii_letters + string.digits))
    def test_make_cid(self, hash):
        is_cid(hash)

    @pytest.mark.parametrize(
        "test_cidv1",
        (
            "bafybeigdyrzt5sfp7udm7hu76uh7y26nf3efuylqabf3oclgtqy55fbzdi",
            "bafkreigh2akiscaildcqabsyg3dfr6chu3fgpregiymsck7e7aqa4s52zy",
            "zb2rhk6GMPQF3hfzwXTaNYFLKomMeC6UXdUt6jZKPpeVirLtV",
        ),
    )
    def test_is_cidv1_valid(self, test_cidv1):
        assert is_cid(test_cidv1)
        assert is_cid(make_cid(test_cidv1).encode())

    @pytest.mark.parametrize(
        "test_data",
        (
            "!",
            "a",
            "1",
            "foobar",
            b"foobar",
            multibase.encode("base58btc", b"foobar"),
        ),
    )
    def test_is_cid_invalid(self, test_data):
        assert not is_cid(test_data)


class MakeCIDTestCase(object):
    def test_base_encoded_hash(self, test_hash):
        """make_cid: make_cid works with base-encoded hash"""
        assert make_cid(base58.b58encode(test_hash)) == CIDv0(test_hash)

    def test_multibase_hash(self, test_hash):
        """make_cid: make_cid works with multibase-encoded hash"""
        cidstr = CIDv1("dag-pb", test_hash).encode()
        assert make_cid(cidstr) == CIDv1("dag-pb", test_hash)

    @pytest.mark.parametrize(
        "value",
        (
            1,
            1.0,
            object(),
            3 + 2j,
            [],
        ),
    )
    def test_hash_invalid_type(self, value):
        """make_cid: make_cid does not work if first argument is not a str or byte"""
        with pytest.raises(ValueError) as excinfo:
            make_cid(value)
        assert "expected: str or byte" in str(excinfo.value)

    @pytest.mark.parametrize("version", list(range(2, 5)))
    def test_version_invalid(self, version):
        """make_cid: make_cid does not work if version is not 0 or 1"""
        with pytest.raises(ValueError) as excinfo:
            make_cid(version, "dag-pb", b"multihash")
        assert "version should be 0 or 1" in str(excinfo.value)

    def test_codec_invalid(self):
        """make_cid: make_cid does not work if codec is invalid"""
        with pytest.raises(ValueError) as excinfo:
            make_cid(1, "some-random-codec", b"multihash")
        assert "invalid codec" in str(excinfo.value)

    @pytest.mark.parametrize(
        "value",
        (
            1,
            1.0,
            object(),
            3 + 2j,
            [],
        ),
    )
    def test_multihash_invalid(self, value):
        """make_cid: make_cid does not work if multihash type is invalid"""
        with pytest.raises(ValueError) as excinfo:
            make_cid(1, "dag-pb", value)
        assert "invalid type for multihash" in str(excinfo.value)

    def test_version_0_invalid_codec(self):
        """make_cid: make_cid does not work if version 0 has incorrect codec"""
        with pytest.raises(ValueError) as excinfo:
            make_cid(0, "raw", b"multihash")
        assert "codec for version 0" in str(excinfo.value)

    def test_invalid_arguments(self):
        """make_cid: make_cid does not work if bad arguments are passed"""
        with pytest.raises(ValueError) as excinfo:
            make_cid(1, 2, 3, 4)
        assert "invalid number of arguments" in str(excinfo.value)


class FromStringTestCase(object):
    @pytest.fixture()
    def cidv0(self, test_hash):
        return CIDv0(test_hash)

    @pytest.fixture()
    def cidv1(self, test_hash):
        return CIDv1("dag-pb", test_hash)

    @pytest.mark.parametrize("codec", ALLOWED_ENCODINGS)
    def test_multibase_encoded_hash(self, cidv1, codec):
        """from_string: works for multibase-encoded strings"""
        assert from_string(cidv1.encode(codec.encoding)) == cidv1

    def test_cid_raw(self, cidv0, cidv1):
        """from_string: works for raw cidbytes"""
        assert from_string(cidv1.buffer) == cidv1

    def test_base58_encoded_hash(self, cidv0):
        """from_string: works for base58-encoded strings"""
        assert from_string(cidv0.encode()) == cidv0

    def test_invalid_base58_encoded_hash(self):
        with pytest.raises(ValueError) as excinfo:
            from_string("!!!!")
        assert "multihash is not a valid base58 encoded multihash" in str(excinfo.value)

    @pytest.mark.parametrize("value", ("", "a"))
    def test_invalid_length_zero(self, value):
        with pytest.raises(ValueError) as excinfo:
            from_string(value)
        assert "argument length can not be zero" in str(excinfo.value)

    def test_invalid_cid_length(self):
        with pytest.raises(ValueError) as excinfo:
            from_string("011111111")
        assert "cid length is invalid" in str(excinfo.value)
